/*! ********************************************
 * SynthesizerWT.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth
{
import flash.utils.ByteArray;
import sora.graybox.MidiEvent;
// WaveTable Synthesis
public final class SynthesizerWT extends MultiTimberSynthBase
{
// Factory -----
	private function _resetIndexTable():void
	{
		var _size:int = 1 << _maxWaveValueBits, _half:int = _size >> 1;
		if (_indexTable == null)_indexTable = new Vector.<int>(_size, true);
		const _LOGE:Number = -(SynthData.LOG2VT_DYN_RES * 1.442695040888963387),
		      _BOADER:Number = 1.0 / (1 << SynthData.LOG2VT_DYN_BITS),
		      _BOTTOM:int = SynthData.LOG2VT_SIZE;
		var i:int, v:Number;
		for (i = 0; i < _size; i++)
		{
			v = (i - _half) * 0.0078125;
			if      (v >  _BOADER) _indexTable[i] =  int(Math.log( v) * _LOGE + 0.5) << 1;
			else if (v < -_BOADER) _indexTable[i] = (int(Math.log(-v) * _LOGE + 0.5) << 1) + 1;
			else _indexTable[i] = _BOTTOM;
		}
	}
	override protected function _recreateVU(maxPoly:int):void
	{
		// Recreate Array
		if (_voiceUnits == null)_voiceUnits = new Vector.<VoiceUnitBase>(maxPoly, true);
		else 
		{
			_voiceUnits.fixed = false;
			_voiceUnits.length = maxPoly;
			_voiceUnits.fixed = true;
		}
		if (_voiceUnitExt == null)_voiceUnitExt = new Vector.<VoiceUnitWT>(maxPoly, true);
		else 
		{
			_voiceUnitExt.fixed = false;
			_voiceUnitExt.length = maxPoly;
			_voiceUnitExt.fixed = true;
		}
		// Reset VoiceUnit
		var i:int, _vu:VoiceUnitWT;
		for (i = 0; i < maxPoly; i++)
		{
			if (_voiceUnitExt[i] == null)
			{
				_voiceUnitExt[i] = new VoiceUnitWT();
				_voiceUnitExt[i].setup(_synthData);
			}
			_voiceUnits[i] = _vu = _voiceUnitExt[i];
			_vu.killProcess();
			_vu.VOICE_PARTS = voiceParts;
			_vu.WAVE_TABLES = waveTables;
			_vu.TONE_PRESETS = tonePresets;
			_vu.SAMPLE_DATAS = sampleDatas;
			_vu.SAMPLER_PRESETS = samplerPresets;
			//_vu.waveTableSize = _waveTableSize;
			_vu.pgPhaseMask = (_waveTableSize << PG_PHASE_FP) - 1;
			_vu.pgPhaseMaskNoise = _vu.pgPhaseMask >> 5;  // 32/1 update cycle
			_vu.setMasterVolume(_masterVolume);
		}
		// set Property
		_maxPolyphony = maxPoly;
	}
	override protected function _resetKeyParams():void
	{
		// LFO
		var _lfoWaveSize:Number = SynthData.LFO_WAVE_TABLE_SIZE / 2.0,
			// Max Frame Rate - LFO Phase Step
		    _lfoMaxFrameRate:Number = _sampleRate / SynthData.LFO_MAX_FREQ / _lfoWaveSize;
		if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
		var _lfoTimerCoef:int =  int(SynthData.LFO_TIMER_INIT_VALUE / 16383.0 / _lfoMaxFrameRate + 0.5);
		// EG
		var _rate0Frames:Number = SynthData.EG_REFERENCE_TIME * SynthData.LOG2VT_DYN_BITS * _sampleRate,  // to -96dB Frames
			_egPow2tIdx:int = SynthData.toPow2tIndexValue(SynthData.EG_MAX_CNT / _rate0Frames),
			_sndOffFrames:Number = SynthData.EG_SOUND_OFF_TIME / SynthData.LOG2VT_DYN_BITS * _sampleRate,  // to -96dB Frames
			_egSndOff:int = int(SynthData.EG_MAX_CNT / _sndOffFrames + 0.5);
		// PG
		var _smpl:int = SynthData.toPow2tIndexValue(_sampleRate);
		var _pgSamplePow2tIdx:int = SynthData.toPow2tIndexValue(1 << PG_PHASE_FP) - _smpl;
		// PG WaveTable
		var _pgKeyPow2tIdx:int = SynthData.toPow2tIndexValue(_waveTableSize << PG_PHASE_FP) - _smpl;
		_pgKeyPow2tIdx += SynthData.toPow2tIndexValue(_referencePitch * Math.pow(2.0, -69.0 / 12.0)); // note 0 Pitch
		var i:int, _vu:VoiceUnitWT;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnitExt[i];
			_vu.killProcess();
			_vu.pgSamplePow2tIdx = _pgSamplePow2tIdx;
			_vu.pgKeyPow2tIdx = _pgKeyPow2tIdx;
			_vu.egPow2tIdx = _egPow2tIdx;
			_vu.egPhaseStepSndOff = _egSndOff;
			_vu.lfoTimerCoef = _lfoTimerCoef;
		}
	}
// Tables -----
	private var _indexTable:Vector.<int> = null;
// Synthesizer - Data -----
	static internal const WAVET_CONT_SIZE_BITS:int = 7;
	static internal const WAVET_CONT_SIZE     :int = 1 << WAVET_CONT_SIZE_BITS;
	static internal const PG_PHASE_FP :int = 18;  // 18bit fixed-point  (precision 1 / 262144)
	private var waveTables  :Vector.<Vector.<int>> = null; // indexed Values
	private var sampleDatas :Vector.<SampleData> = null;
	public var tonePresets    :Vector.<TonePresetWT> = null;  // [0 - 127](ProgramChange), [128 - 255](Percussion Key)
	public var samplerPresets :Vector.<SamplerPreset> = null;
	private var _waveTableSizeBits:int = 5;
	private var _waveTableSize    :int = 1 << _waveTableSizeBits;
	private var _maxWaveValueBits :int = 8;
	private var _synthData:SynthData = null;
	public function getWaveTableSize():int { return _waveTableSize; }
	public function setWaveTableSize(size:int):int
	{
		if (size == _waveTableSize) return _waveTableSize;
		var v:int = 8, _b:int = 3, _start:int = _waveTableSize;
		for ( ; v < WAVET_CONT_SIZE; v <<= 1, _b++) if (v >= size) break;
		_waveTableSizeBits = _b;
		_waveTableSize = v;
		var i:int, j:int, _table:Vector.<int>;
		const _BOTTOM:int = SynthData.LOG2VT_SIZE;
		for (i = 0; i < 128; i++)
		{
			_table = waveTables[i];
			for (j = _start; j < _waveTableSize; j++)_table[j] = _BOTTOM;
		}
		var _vu:VoiceUnitWT;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnitExt[i];
			_vu.killProcess();
			//_vu.waveTableSize = _waveTableSize;
			_vu.pgPhaseMask = (_waveTableSize << PG_PHASE_FP) - 1;
			_vu.pgPhaseMaskNoise = _vu.pgPhaseMask >> 5;
		}
		_resetKeyParams();
		return _waveTableSize;
	}
	public function getWaveForm(id:int, buffer:Vector.<int>, bitDepth:int):void
	{
		var _size:int = buffer.length;
		if (_size > WAVET_CONT_SIZE)_size = WAVET_CONT_SIZE;
		// ---
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		const _center:Number = 1 << (bitDepth - 1);
		const _max:int = (1 << bitDepth) - 1;
		var i:int, v:int, _table:Vector.<int> = waveTables[id];
		const _log2vt:Vector.<Number> = _synthData.getLOG2V_TABLE();
		for (i = 0; i < _size; i++)
		{
			v = int(_log2vt[_table[i]] * _center + _center + 0.5);
			if (v > _max) v = _max;
			buffer[i] = v;
		}
	}
	public function getWaveFormN(id:int, buffer:Vector.<Number>):void
	{
		var _size:int = buffer.length;
		if (_size > WAVET_CONT_SIZE)_size = WAVET_CONT_SIZE;
		// ---
		var i:int, _table:Vector.<int> = waveTables[id];
		const _log2vt:Vector.<Number> = _synthData.getLOG2V_TABLE();
		for (i = 0; i < _size; i++) buffer[i] = _log2vt[_table[i]];
	}
	public function getWaveTableValue(id:int, index:int, bitDepth:int):int
	{
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		const _center:Number = 1 << (bitDepth - 1);
		const _max:int = (1 << bitDepth) - 1;
		var v:int = int(_synthData.getLOG2V_TABLE()[waveTables[id][index]] * _center + _center + 0.5);
		if (v > _max) v = _max;
		return v;
	}
	public function getWaveTableValueN(id:int, index:int):Number
	{
		return _synthData.getLOG2V_TABLE()[waveTables[id][index]];
	}
	// buffer [0-255(128 = 0)](bitDepth = 8)
	public function setWaveForm(id:int, buffer:Vector.<int>, bitDepth:int):void
	{
		var _size:int = buffer.length;
		if (_size > WAVET_CONT_SIZE)_size = WAVET_CONT_SIZE;
		// ---
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		const _shift:int = 8 - bitDepth;
		var i:int, v:int, _table:Vector.<int> = waveTables[id];
		for (i = 0; i < _size; i++)
		{
			v = buffer[i] << _shift;
			_table[i] = _indexTable[v & 0xff];
		}
	}
	// buffer [-1.0 ~ 1.0]
	public function setWaveFormN(id:int, buffer:Vector.<Number>, bitDepth:int = 8):void
	{
		var _size:int = buffer.length;
		if (_size > WAVET_CONT_SIZE)_size = WAVET_CONT_SIZE;
		// ---
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		const _shift:int = 8 - bitDepth;
		var i:int, v:int, _table:Vector.<int> = waveTables[id];
		for (i = 0; i < _size; i++)
		{
			v = int(buffer[i] * 127.0) + 128;
			v >>= _shift;
			_table[i] = _indexTable[v];
		}
	}
	public function setWaveTableValue(id:int, index:int, value:int, bitDepth:int):void
	{
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		waveTables[id][index] = _indexTable[value << (8 - bitDepth) & 0xff];
	}
	public function setWaveTableValueN(id:int, index:int, value:Number, bitDepth:int = 8):void
	{
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		var v:int = int(value * 127.0) + 128;
		waveTables[id][index] = _indexTable[v >> (8 - bitDepth)];
	}
	public function getSampleData(id:int):SampleData{ return sampleDatas[id]; }
	public function setSampleData(id:int, sampleData:SampleData, deepCopy:Boolean = true):void
	{
		allSoundStop();
		sampleDatas[id].copy(sampleData, deepCopy);
	}
	public function loadPCMSamples(id:int, wave:Vector.<Number>, ch:int):void
	{
		allSoundStop();
		sampleDatas[id].loadPCMSamples(wave, ch);
	}
	public function loadPCMData(id:int, dat:ByteArray, sampleFrames:int, ch:int, bitRate:int, sign:Boolean, bigEndian:Boolean):void
	{
		allSoundStop();
		sampleDatas[id].loadPCMData(dat, sampleFrames, ch, bitRate, sign, bigEndian);
	}
	public function clearSampleData(id:int):void
	{
		allSoundStop();
		sampleDatas[id].clear();
	}
	override public function factoryDataReset():void
	{
		if (_voiceUnitExt != null) allSoundStop();
		// waveData
		const _PID2:Number = 3.141592653589793 * 2.0;
		var i:int, j:int, _nv:Number, _iv:int;
		var _simpleWave:Vector.<int> = new Vector.<int>(WAVET_CONT_SIZE, true);
		for (i = 0; i < _waveTableSize; i++)
		{
			_nv = Math.sin((i / _waveTableSize) * _PID2);
			_iv = int(_nv * 127.5 - 0.5) + 128;
			_simpleWave[i] = _indexTable[_iv];
		}
		if (waveTables == null)waveTables = new Vector.<Vector.<int>>(128, true);
		for (i = 0; i < 128; i++)
		{
			if (waveTables[i] == null)waveTables[i] = new Vector.<int>(WAVET_CONT_SIZE, true);
			for (j = 0; j < _waveTableSize; j++) waveTables[i][j] = _simpleWave[j];
		}
		// Preset
		if (tonePresets == null)tonePresets = new Vector.<TonePresetWT>(256, true);
		for (i = 0; i < 256; i++)
		{
			if (tonePresets[i] == null)tonePresets[i] = new TonePresetWT();
			else tonePresets[i].reset();
		}
	// Sampler
		if (sampleDatas == null)sampleDatas = new Vector.<SampleData>(128, true);
		if (samplerPresets == null)samplerPresets = new Vector.<SamplerPreset>(128, true);
		for (i = 0; i < 128; i++)
		{
			if (sampleDatas[i] == null)sampleDatas[i] = new SampleData();
			else sampleDatas[i].clear();
			if (samplerPresets[i] == null)samplerPresets[i] = new SamplerPreset();
			else samplerPresets[i].reset();
		}
	}
// Synthesizer - main -----
	private var _voiceUnitExt :Vector.<VoiceUnitWT> = null;
// Constructor
	public function SynthesizerWT(tblSize:int = 32, maxPoly:int = 32, refPitch:Number = 440.0, masVol:Number = 1.0, privateTableData:Boolean = true):void
	{
	// super ===
		super(maxPoly, refPitch, masVol);
	// SynthesizerWT ===
		_modelID = 0x20;
		_maxPMN = 3;
		var v:int, _b:int;
		for (v = 8, _b = 3; v < WAVET_CONT_SIZE; v <<= 1, _b++) if (v >= tblSize) break;
		_waveTableSizeBits = _b;
		_waveTableSize = v;
		// create
		_synthData = (privateTableData)? new SynthData() : SynthData.getInstance();
		_resetIndexTable();
		factoryDataReset();
		_recreateVU(_maxPolyphony);
		_resetKeyParams();
	}
// MidiEvent ---
	override protected function _noteOn(ev:MidiEvent):void
	{
		var _vu:VoiceUnitWT, _pt:VoicePart = voiceParts[ev.channel];
		const _start:int = _nton;
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			_vu = _voiceUnitExt[_nton];
			if (_vu.state == 0)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
			
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			_vu = _voiceUnitExt[_nton];
			// (KeyOff || SoundOff) -->
			if (_vu.state & 0x60)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}

		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			_vu = _voiceUnitExt[_nton];
			// KeyOn -->
			if (_vu.state & 0x10)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
	}
	override protected function _readSysExData(ev:MidiEvent):void
	{
		var v:int = 0, _data:ByteArray = ev.data;
		_data.position = 0;
		if (_data.readByte() != _manufacID) return;
		if (_data.readByte() != _deiceID) return;
		if (_data.readByte() != _modelID) return;
		switch (_data.readByte()) // Address 1
		{
		// Common
			case 0x00:
				switch (_data.readUnsignedShort())	// Address 2,3
				{
					case 0x0000 : setMasterVolume(_masterVolumeTable[_data.readByte()]);     if (_data.bytesAvailable <= 1) break;
					case 0x0001 : setReferencePitch(440.0 + (_data.readByte() - 64.0) * 0.5);  break;
				}
				break;
		// Part
			case 0x02:
				v = _data.readByte() & 15;	// Address 2 [ CH ]
				switch (_data.readByte())	// Address 3
				{
					case 0x00 : setDefaultPartMode(v, _data.readByte()); if (_data.bytesAvailable <= 1) break;
					case 0x01 : setPartMode(v, _data.readByte()); break;
				}
				break;
		// Inst (Drum map)
			case 0x05: v = 128;
		// Inst (Program map)
			case 0x04:
				var _preset:TonePresetWT = tonePresets[v | _data.readByte()];	// Address 2 [ ID ]
				switch (_data.readByte())	// Address 3
				{
					case 0x00 : _preset.toneName = _data.readMultiByte(16, 'us-ascii');             if (_data.bytesAvailable <= 1) break;
					case 0x10 : _preset.key = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x11 : _preset.pan = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x12 : _preset.lfoFreq = _preset.lfoFreq & 0x3f80 | _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x13 : _preset.lfoFreq = _preset.lfoFreq & 0x007f | _data.readByte() << 7; if (_data.bytesAvailable <= 1) break;
					case 0x14 : _preset.strc = _data.readByte() & 3; if (_data.bytesAvailable <= 1) break;
					case 0x15 : _preset.tune = _preset.tune & 0x3f80 | _data.readByte();            if (_data.bytesAvailable <= 1) break;
					case 0x16 : _preset.tune = _preset.tune & 0x007f | _data.readByte() << 7;       if (_data.bytesAvailable <= 1) break;
					case 0x17 : _preset.lms = _data.readByte() & 15; if (_data.bytesAvailable <= 1) break;
					case 0x18 : _preset.wave = _data.readByte();     if (_data.bytesAvailable <= 1) break;
					case 0x19 : _preset.tl = _data.readByte();       if (_data.bytesAvailable <= 1) break;
					case 0x1A : _preset.ar = _data.readByte();       if (_data.bytesAvailable <= 1) break;
					case 0x1B : _preset.dr = _data.readByte();       if (_data.bytesAvailable <= 1) break;
					case 0x1C : _preset.sl = _data.readByte();       if (_data.bytesAvailable <= 1) break;
					case 0x1D : _preset.rr = _data.readByte();       if (_data.bytesAvailable <= 1) break;
					case 0x1E : _preset.kf = _data.readByte() & 3;   if (_data.bytesAvailable <= 1) break;
					case 0x1F : _preset.kfStart = _data.readByte();  break;
				}
				break;
		// Percussion
			case 0x06:
				var  _smplPrst:SamplerPreset = samplerPresets[_data.readByte()];	// Address 2 [ ID ]
				switch (_data.readByte())	// Address 3
				{
					case 0x00 : _smplPrst.toneName = _data.readMultiByte(16, 'us-ascii');              if (_data.bytesAvailable <= 1) break;
					case 0x10 : _smplPrst.smplNum = _data.readByte();   if (_data.bytesAvailable <= 1) break;
					case 0x11 : _smplPrst.pan = _data.readByte();       if (_data.bytesAvailable <= 1) break;
					case 0x12 : _smplPrst.octv = _data.readByte() & 15;	if (_data.bytesAvailable <= 1) break;
					case 0x13 : _smplPrst.tune = _smplPrst.tune & 0x3f80 | _data.readByte();           if (_data.bytesAvailable <= 1) break;
					case 0x14 : _smplPrst.tune = _smplPrst.tune & 0x007f | _data.readByte() << 7;      if (_data.bytesAvailable <= 1) break;
					case 0x15 : _smplPrst.tl = _data.readByte();        if (_data.bytesAvailable <= 1) break;
					case 0x16 : _smplPrst.ar = _data.readByte();        if (_data.bytesAvailable <= 1) break;
					case 0x17 : _smplPrst.dr = _data.readByte();        if (_data.bytesAvailable <= 1) break;
					case 0x18 : _smplPrst.sl = _data.readByte();        if (_data.bytesAvailable <= 1) break;
					case 0x19 : _smplPrst.rr = _data.readByte();        break;
				}
				break;
		// Data
			case 0x08:
				v = _data.readByte() & 127;	// Address 2 [ ID ]
				for (var i:int = _data.readByte(); i < WAVET_CONT_SIZE; i++)
				{
					waveTables[v][i] = _indexTable[_data.readByte() << 1];
					if (_data.bytesAvailable <= 1) break;
				}
				break;
		}
	}

}
}