/*! ********************************************
 * TonePresetFM4.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth
{
/*
	- Parameters Notes -
	
	[toneName]
	  16 Chars (ASCII Code)
	[key] * This parameter is enabled when the VoicePart is in mode 2
	  Key Pitch [NoteNumber(0~127)]
	
	[pan] * Enabled -> VoicePart mode 2
	  Panpot [ 0(L) ~ 64(center) ~ 127(R) ]
	
	[lfoWave]
	  LFO WaveForm [ 0 ~ 3 ]
	      0 -> Triangle Wave
	      1 -> Square Wave
	      2 -> Saw Wave
	      3 -> Noise
	  
	[lfoFreq]
	  LFO Frequency [ 0 ~ 16383 ]
	  ( Frequency = lfoFreq * 0.001953125 )
	      0 ->  0 Hz
	    128 ->  0.25 Hz
	    256 ->  0.50 Hz
	    512 ->  1.00 Hz
	   1024 ->  2.00 Hz
	  16383 -> 31.998046875 Hz
	
	[fl]
	  Feedback Level [ 0 ~ 127 ]
	    0 ->  0dB
	    4 -> -3dB
	    8 -> -6dB
	   16 -> -12dB
	  127 -> -96dB
	
	[strc]
	Structure Number (Algorithm) [ 0 ~ 15 ] ([0~7] -> LFO:PM )([8~15] -> LFO:AM )
	----------------------------------------------------------------------
	0( 8)| [OP0 FeedBack] -> [OP 0] -|
		 |                   [OP 1] -+-> out
		 |                   [OP 2] -|
		 |                   [OP 3] -|
	----------------------------------------------------------------------
	1( 9)| [OP0 FeedBack] -> [OP 0] -> [OP 1] -|
		 |                             [OP 2] -+-> out
		 |                             [OP 3] -|
	----------------------------------------------------------------------
	2(10)|                           |-> [OP 1] -|
		 | [OP0 FeedBack] -> [OP 0] -+-> [OP 2] -+-> out
		 |                           |-> [OP 3] -|
	----------------------------------------------------------------------
	3(11)| [OP0 FeedBack] -> [OP 0] -> [OP 1] -+-> out
		 |                   [OP 2] -> [OP 3] -|
	----------------------------------------------------------------------
	4(12)| [OP0 FeedBack] -> [OP 0] -> [OP 1] -+-> [OP 3] -> out
		 |                             [OP 2] -|
	----------------------------------------------------------------------
	5(13)| [OP0 FeedBack] -> [OP 0] -+-> [OP 3] -> out
		 |         [OP 1] -> [OP 2] -| 
	----------------------------------------------------------------------
	6(14)|  [OP0 FeedBack] -> [OP 0] -+-> [OP 2] -> [OP 3] -> out
		 |                    [OP 1] -|
	----------------------------------------------------------------------
	7(15)|  [OP 0 FeedBack] -> [OP 0] -> [OP 1] -> [OP 2] -> [OP 3] -> out
	----------------------------------------------------------------------	
	  
	[tlVelCtrl]
	  Total Level Control by Velocity Level [0 ~ 7]
	  TotalLevel = tl[0 ~ 127] + ((127 - Velocity[0 ~ 127]) >> (7 - tlVelCtrl[0 ~ 7]))
	  TotalLevel = (TotalLevel > 127)? 127 : TotalLevel;
	  
	[mul]
	  Multiple [0 ~ 15]
	  0 -> 1/2 pai
	  1 -> 1/1 pai
	  2 ->  2  pai
	  3 ->  3  pai
	  7 ->  7  pai
	  
	[tune]
	  Tune [0(-1 octave) ~ 8192(+-0) ~ 16383(+1 octave)]
	  8192 / 1 octave
	  
	[lms]
	  LFO Modulation Depth Senstivity [0 ~ 15]
	  Modulation Depth = Midi control Change 1[0~127] * lms[0 ~ 15]
	  
	[wave]
	  WaveTable Index Nuber [0 ~ 7]
	  
	[tl]
	  Total Level [0 ~ 127]
	  0 -> 0dB
	  4 -> -3dB
	  8 -> -6dB
	  16 -> -12dB
	  127 -> -96dB
	  
	[ar]
	  Attack Rate [0 ~ 127]
	    0 -> AttackTime 256.0Sec / 1
	    8 -> AttackTime 256.0Sec / 2
	   16 -> AttackTime 256.0Sec / 4
	   24 -> AttackTime 256.0Sec / 8
	   32 -> AttackTime 256.0Sec / 16
	  128 -> AttackTime 256.0Sec / 65536
	  
	[dr]
	  Decay Rate [0 ~ 127]
	    0 -> DecayTime 256.0Sec / 1
	    8 -> DecayTime 256.0Sec / 2
	   16 -> DecayTime 256.0Sec / 4
	   24 -> DecayTime 256.0Sec / 8
	   32 -> DecayTime 256.0Sec / 16
	  128 -> DecayTime 256.0Sec / 65536
	  
	[sl]
	  Sustain Level [0 ~ 127]
	  0 -> 0dB
	  4 -> -3dB
	  8 -> -6dB
	  16 -> -12dB
	  127 -> -96dB
	  
	[rr]
	  Release Rate [0 ~ 127]
	    0 -> DecayTime 256.0Sec / 1
	    8 -> DecayTime 256.0Sec / 2
	   16 -> DecayTime 256.0Sec / 4
	   24 -> DecayTime 256.0Sec / 8
	   32 -> DecayTime 256.0Sec / 16
	  128 -> DecayTime 256.0Sec / 65536
	  
	[kf] * Enabled -> VoicePart mode 1
	  Key Follow Sensitivity [0 - 3] (amp EG)
	  0 -> 48 Key / 1 octave
	  1 -> 24 Key / 1 octave
	  2 -> 12 Key / 1 octave
	  3 ->  6 Key / 1 octave
	  
	[kfStart] * Enabled -> VoicePart mode 1
	  StartKey [0 - 127] (Key Follow)
	
	-------------------------
	VoicePart Mode  | 1 | 2 |
	-------------------------
	[key]           | x | o |
	[pan]           | x | o |
	[lfoWave]       | o | o |
	[lfoFreq]       | o | o |
	[fl]            | o | o |
	[strc]          | o | o |
	[tlVelCtrl]     | o | o |
	[mul]           | o | o |
	[tune]          | o | o |
	[lms]           | o | o |
	[wave]          | o | o |
	[tl]            | o | o |
	[ar]            | o | o |
	[dr]            | o | o |
	[sl]            | o | o |
	[rr]            | o | o |
	[kf]            | o | x |
	[kfStart]       | o | x |
	-------------------------
	o = parameter is enable.
	x = parameter is disable.
	-------------------------
	VoicePart Mode 0 -> Mute (channel mute)
	VoicePart Mode 1 -> Instrument Mode (Assign TonePreset[  0 ~ 127] to NoteNumber[0 ~ 127] )
	VoicePart Mode 2 -> Percussion Mode (Assign TonePreset[128 ~ 255] to NoteNumber[0 ~ 127] )
	VoicePart Mode 3 -> Sampler Mode
	-------------------------
*/
public class TonePresetFM4
{
// Inastance -----
	public var toneName		:String = "  default preset";
	public var key			:int = 64;   // 0 - 127     *Percussion
	public var pan			:int = 64;   // 0 - 127     *Percussion
	public var lfoWave		:int = 0;    // 0 - 3
	public var lfoFreq		:int = 2048; // 0 - 16383   LFO - Phase Update Timer Parameter
	public var fl			:int = 127;  // 0 - 127     Feed Back Level
	public var strc			:int = 0;    // 0 - 15	    strcture
	// operator 0
	//public var velCurve_0	:int;
	public var tlVelCtrl_0	:int = 0;   // 0 - 7        Total Level Contorol - Velocity   Senstivity
	public var mul_0		:int = 1;   // 0 - 15        Multiple       [0] = 1/2 Cycle, [1] = 1/1, [2] = 2/1, [3] = 3/1...
	public var tune_0		:int = 8192;// 0 - 16383    center 8192, octave = 8192
	public var lms_0		:int = 0;   // 0 - 15       LFO - Modulation Depth Senstivity
	public var wave_0		:int = 0;   // 0 - 7
	public var tl_0			:int = 0;   // 0 - 127
	public var ar_0			:int = 124; // 0 - 127
	public var dr_0			:int = 100; // 0 - 127
	public var sl_0			:int = 0;   // 0 - 127
	public var rr_0			:int = 120; // 0 - 127
	public var kf_0			:int = 0;   // 0 - 3        Key Follow Sensitivity         *Instrument 
	public var kfStart_0	:int = 48;  // 0 - 127      Key Number - Start Key Follow  *Instrument
	// operator 1
	public var tlVelCtrl_1	:int = 0;
	public var mul_1		:int = 1;
	public var tune_1		:int = 8192;
	public var lms_1		:int = 0;
	public var wave_1		:int = 0;
	public var tl_1			:int = 0;
	public var ar_1			:int = 124;
	public var dr_1			:int = 100;
	public var sl_1			:int = 0;
	public var rr_1			:int = 120;
	public var kf_1			:int = 0;
	public var kfStart_1	:int = 48;
	// operator 2
	public var tlVelCtrl_2	:int = 0;
	public var mul_2		:int = 1;
	public var tune_2		:int = 8192;
	public var lms_2		:int = 0;
	public var wave_2		:int = 0;
	public var tl_2			:int = 0;
	public var ar_2			:int = 124;
	public var dr_2			:int = 100;
	public var sl_2			:int = 0;
	public var rr_2			:int = 120;
	public var kf_2			:int = 0;
	public var kfStart_2	:int = 48;
	// operator 3
	public var tlVelCtrl_3	:int = 0;
	public var mul_3		:int = 1;
	public var tune_3		:int = 8192;
	public var lms_3		:int = 0;
	public var wave_3		:int = 0;
	public var tl_3			:int = 0;
	public var ar_3			:int = 124;
	public var dr_3			:int = 100;
	public var sl_3			:int = 0;
	public var rr_3			:int = 120;
	public var kf_3			:int = 0;
	public var kfStart_3	:int = 48;
// Method
	public function copy(preset:TonePresetFM4):void
	{
		toneName = preset.toneName;
		key = preset.key;
		pan = preset.pan;
		lfoWave = preset.lfoWave;
		lfoFreq = preset.lfoFreq;		
		fl = preset.fl;
		strc = preset.strc;
		// operator 0
		tlVelCtrl_0	= preset.tlVelCtrl_0;
		mul_0		= preset.mul_0;
		tune_0		= preset.tune_0;
		lms_0		= preset.lms_0;
		wave_0		= preset.wave_0;
		tl_0		= preset.tl_0;
		ar_0		= preset.ar_0;
		dr_0		= preset.dr_0;
		sl_0		= preset.sl_0;
		rr_0		= preset.rr_0;
		kf_0		= preset.kf_0;
		kfStart_0	= preset.kfStart_0;
		// operator 1
		tlVelCtrl_1	= preset.tlVelCtrl_1;
		mul_1		= preset.mul_1;
		tune_1		= preset.tune_1;
		lms_1		= preset.lms_1;
		wave_1		= preset.wave_1;
		tl_1		= preset.tl_1;
		ar_1		= preset.ar_1;
		dr_1		= preset.dr_1;
		sl_1		= preset.sl_1;
		rr_1		= preset.rr_1;
		kf_1		= preset.kf_1;
		kfStart_1	= preset.kfStart_1;
		// operator 2
		tlVelCtrl_2	= preset.tlVelCtrl_2;
		mul_2		= preset.mul_2;
		tune_2		= preset.tune_2;
		lms_2		= preset.lms_2;
		wave_2		= preset.wave_2;
		tl_2		= preset.tl_2;
		ar_2		= preset.ar_2;
		dr_2		= preset.dr_2;
		sl_2		= preset.sl_2;
		rr_2		= preset.rr_2;
		kf_2		= preset.kf_2;
		kfStart_2	= preset.kfStart_2;
		// operator 3
		tlVelCtrl_3	= preset.tlVelCtrl_3;
		mul_3		= preset.mul_3;
		tune_3		= preset.tune_3;
		lms_3		= preset.lms_3;
		wave_3		= preset.wave_3;
		tl_3		= preset.tl_3;
		ar_3		= preset.ar_3;
		dr_3		= preset.dr_3;
		sl_3		= preset.sl_3;
		rr_3		= preset.rr_3;
		kf_3		= preset.kf_3;
		kfStart_3	= preset.kfStart_3;
	}
	public function reset():void
	{
		toneName = "  default preset";
		key = 64;
		pan = 64;
		lfoWave = 0;
		lfoFreq = 2048;
		fl = 127;
		strc = 0;
		// operator 0
		tlVelCtrl_0	= 0;
		mul_0		= 1;
		tune_0		= 8192;
		lms_0 		= 0;
		wave_0		= 0;
		tl_0		= 0;
		ar_0		= 124;
		dr_0		= 100;
		sl_0		= 0;
		rr_0		= 120;
		kf_0		= 0;
		kfStart_0	= 48;
		// operator 1
		tlVelCtrl_1	= 0;
		mul_1		= 1;
		tune_1		= 8192;
		lms_1 		= 0;
		wave_1		= 0;
		tl_1		= 0;
		ar_1		= 124;
		dr_1		= 100;
		sl_1		= 0;
		rr_1		= 120;
		kf_1		= 0;
		kfStart_1	= 48;
		// operator 2
		tlVelCtrl_2	= 0;
		mul_2		= 1;
		tune_2		= 8192;
		lms_2 		= 0;
		wave_2		= 0;
		tl_2		= 0;
		ar_2		= 124;
		dr_2		= 100;
		sl_2		= 0;
		rr_2		= 120;
		kf_2		= 0;
		kfStart_2	= 48;
		// operator 3
		tlVelCtrl_3	= 0;
		mul_3		= 1;
		tune_3		= 8192;
		lms_3 		= 0;
		wave_3		= 0;
		tl_3		= 0;
		ar_3		= 124;
		dr_3		= 100;
		sl_3		= 0;
		rr_3		= 120;
		kf_3		= 0;
		kfStart_3	= 48;
	}
// Constructor
	public function TonePresetFM4():void{}
}
}