/*! ********************************************
 * NoiseGenerator.as
 * Copyright 2022 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> license.txt
 ******************************************** */
package {
	import sora.graybox.GrayBox;
	import sora.graybox.IGrayBoxModule;
	import sora.graybox.ProcessData;
// ノイズジェネレーターモジュール
public class NoiseGenerator implements IGrayBoxModule {
	public var volume:Number = 0.1;  // 音量
	// コンストラクター
	public function NoiseGenerator():void {}
// sora.graybox.IGrayBoxModule -->
	// ホストのサンプリングレートが変更された時に呼ばれるメソッド
	// GrayBox.setSampleRate() -> このメソッド
	public function updateSampleRate(sampleRate:Number):void{}
	// ホストのオーディオチャンネル数が変更された時に呼ばれるメソッド
	// GrayBox.setAudioChannels() -> このメソッド
	public function updateAudioChannels(ch:int):void{}
	// ホストのプロセスブロックサイズが変更された時に呼ばれるメソッド
	// GrayBox.setProcessBlockSize() -> このメソッド
	public function updateProcessBlockSize(size:int):void{}
	// ホストのGrayBox.resetModules()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.resetModules() -> このメソッド
	public function resetModule():void{ volume = 0.1; }
	// ホストのGrayBox.initiModules()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.initiModules() -> このメソッド
	public function initiModule():void{ volume = 0.1; }
	// このモジュールがホストへ接続された時に呼ばれるメソッド
	// GrayBox.attachModule() -> このメソッド
	public function connectModule(host:GrayBox):void{}
	// このモジュールがホストから取り外された時に呼ばれるメソッド
	// GrayBox.detachModule() -> このメソッド
	public function disconnectModule():void{}
	// このモジュールが接続されているスロットの、バイパス状況が更新された時に呼ばれるメソッド
	// GrayBox.bypassProcess() -> このメソッド
	public function updateBypassStatus(bypass:Boolean):void{}
	// ホストのGrayBox.startProcess()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.startProcess() -> このメソッド
	public function startProcess():void{}
	// ホストのGrayBox.stopProcess()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.stopProcess() -> このメソッド
	public function stopProcess():void{}
	// ホストのGrayBox.suspendProcess()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.suspendProcess() -> このメソッド
	public function suspendProcess():void{}
	// ホストのGrayBox.resumeProcess()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.resumeProcess() -> このメソッド
	public function resumeProcess():void{}
	
	// ホストが音声処理を実行した時に呼ばれるメソッド
	// GrayBox.process() -> このメソッド
	public function process(procData:ProcessData):void {
	// 音声処理準備 -->
		// ProcessDataオブジェクトはホストから音声処理に必要な情報を受け取り、
		// 各モジュールへ伝えるオブジェクトです。
		// ここではホストから渡されたオーディオバッファを参照して音声処理行います。
		var _outL:Vector.<Number> = procData.audioOut[0],  // 音声出力L
		    _outR:Vector.<Number> = procData.audioOut[1],  // 音声出力R
			_startFrame:int = procData.offsetFrames,  // 音声処理開始フレーム
			_endFrame:int = _startFrame + procData.sampleFrames;  // 終了フレーム
	// 音声処理 -->
		for (var i:int = _startFrame; i < _endFrame; i++) {
			var _n:Number = Math.random() * volume; // ノイズ音の生成
			// オーディオバッファに対して加算代入
			_outL[i] += _n; 
			_outR[i] += _n;
			// このモジュールしか音声処理を行わない場合は代入だけでも良い
			// _outL[i] = _n; 
			// _outR[i] = _n;
		}
	}
}
}