/*! ********************************************
 * Main.as
 * Copyright 2022 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> license.txt
 ******************************************** */
package {
import flash.display.Sprite;
import flash.events.SampleDataEvent;
import flash.media.Sound;
import flash.utils.ByteArray;
import sora.graybox.GrayBox;

public class Main extends Sprite {
	
	private var _sound:Sound;  // ストリーミング再生を行うオブジェクト
	
	public const AUDIO_CHANNELS:int = 2;  // オーディオチャンネル数
	public const SAMPLE_RATE:Number = 44100.0;  // サンプリングレート
	public const BLOCK_SIZE:int = 4096;   // ブロックサイズ(バッファサイズ)
	public var audioBuffer:Vector.<Vector.<Number>>;  // オーディオバッファ
	public var grayBox:GrayBox;  // GrayBoxシステム基幹オブジェクト(モジュラーホスト)
	public var simpleSeq:SimpleSequencer;  // シンプルシーケンサー(モジュール)
	public var sineGen:SineGenerator;  // サインジェネレーター(モジュール)
	
	// ストリーミング処理 -->
	private function _sampleDataEvent(e:SampleDataEvent):void
	{
	// オーディオバッファのクリア -->
		var _outL:Vector.<Number> = audioBuffer[0],
		    _outR:Vector.<Number> = audioBuffer[1];
		for (var i:int = 0; i < BLOCK_SIZE; i++) {
			_outL[i] = 0.0;
			_outR[i] = 0.0;
		}
	// GrayBoxによる音声処理 -->
		// GrayBox.process()メソッドを使用して音声処理を行います。
		// 第1引数 : 書き込みを行うオーディオバッファ
		// 第2引数 : 書き込み開始フレーム
		// 第3引数 : 書き込みフレーム数
		// ・今回はオーディオバッファ全体を使うので、開始フレームが0、書き込みフレーム数が4096
		// ・スロットナンバー0から順にモジュール側のIGrayBoxModule.process()メソッドが呼ばれます。
		grayBox.process(audioBuffer, 0, BLOCK_SIZE);
	// オーディオバッファからストリームへ書き込み -->
		var _stream:ByteArray = e.data;
		for (i = 0; i < BLOCK_SIZE; i++) {
			_stream.writeFloat(_outL[i]);
			_stream.writeFloat(_outR[i]);
		}
	}
	// コンストラクター
	public function Main():void {
	// エントリーポイント -->
	// ストリーミング再生を行うオブジェクト構築 -->
		_sound = new Sound();
	// オーディオバッファ作成 -->
		audioBuffer = new Vector.<Vector.<Number>>(2, true);
		audioBuffer[0] = new Vector.<Number>(BLOCK_SIZE, true); // チャンネル L
		audioBuffer[1] = new Vector.<Number>(BLOCK_SIZE, true); // チャンネル R
	// GrayBoxの構築 -->
		// 第1引数 : ブロックサイズ         ( 1~ ) FLASH,AIR環境は2048~8192フレーム
		// 第2引数 : サンプリングレート[Hz] ( 1~ ) FLASH,AIR環境は44100Hz固定
		// 第3引数 : オーディオチャンネル数 ( 2~ ) FLASH,AIR環境はLR2チャンネル固定
		// 第4引数 : モジュールスロット数   ( 1~ ) 今回はモジュールが2つなので2
		// ・初期値は必須パラメーターですが、構築時に確定できない場合は暫定値で構いません。
		//   その場合はストリーミングを開始する前に確定させてください。
		//   関連メソッド
		//   ブロックサイズ -> GrayBox.setProcessBlockSize()
		//   サンプリングレート -> GrayBox.setSampleRate()
		//   チャンネル数 -> GrayBox.setAudioChannels()
		//   スロット数 -> GrayBox.setModuleSlots()
		grayBox = new GrayBox(BLOCK_SIZE, SAMPLE_RATE, AUDIO_CHANNELS, 2);
	// 各GrayBoxモジュールの構築 -->
		// 1秒間隔(44100フレーム)でドレミのMidiメッセージを出力し続けるモジュール
		simpleSeq = new SimpleSequencer();
		// Midiメッセージを元にサイン波を出力するモジュール
		sineGen = new SineGenerator();
	// モジュールの接続 -->
		// GrayBox.attachModule()メソッドを使用し、ホストへモジュールを接続します。
		// 第1引数 : 接続するスロットのインデックス
		// 第2引数 : 接続するモジュールオブジェクト
		// 接続されたモジュールは接続スロットの先頭から順に処理が行われます。
		grayBox.attachModule(0, simpleSeq);
		grayBox.attachModule(1, sineGen);
	// ストリーミングの開始 -->
		// ストリーミング前処理
		// ・モジュール側IGrayBoxModule.startProcess()メソッドが呼ばれます。
		//   必要に応じてモジュール側で前処理を行えます。
		grayBox.startProcess();
		// ストリーミング開始
		_sound.addEventListener(SampleDataEvent.SAMPLE_DATA, _sampleDataEvent);
		_sound.play();
	}
}
}